﻿Option Strict On
Option Explicit On

Imports System.Text
Imports System.Drawing.Imaging
Imports System.Runtime.InteropServices

Public Class ShellContextMenu

#Region "API"
    <DllImport("Shell32.dll", EntryPoint:="SHParseDisplayName")>
    <PreserveSig> Private Shared Function SHParseDisplayName(<[In], MarshalAs(UnmanagedType.LPWStr)> pszName As String,
                                                             <[In]> pbc As IntPtr,
                                                             <Out> ByRef ppidl As IntPtr,
                                                             <[In]> sfgaoIn As IntPtr,
                                                             <Out> ByRef psfgaoOut As IntPtr) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHBindToParent")>
    <PreserveSig> Private Shared Function SHBindToParent(<[In]> pidl As IntPtr,
                                                         <[In]> ByRef riid As Guid,
                                                         <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppv As Object,
                                                         <Out> ByRef ppidlLast As IntPtr) As Integer
    End Function

    <DllImport("Shell32.dll", EntryPoint:="SHCreateShellItemArrayFromIDLists")>
    <PreserveSig> Private Shared Function SHCreateShellItemArrayFromIDLists(<[In]> cidl As Integer,
                                                                            <[In]> rgpidl As IntPtr(),
                                                                            <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppsiItemArray As IShellItemArray) As Integer
    End Function

    <DllImport("User32.dll", EntryPoint:="CreatePopupMenu")>
    Private Shared Function CreatePopupMenu() As IntPtr
    End Function

    <DllImport("User32.dll", EntryPoint:="TrackPopupMenuEx")>
    Private Shared Function TrackPopupMenuEx(<[In]> hMenu As IntPtr,
                                             <[In]> uFlags As Integer,
                                             <[In]> x As Integer,
                                             <[In]> y As Integer,
                                             <[In]> hwnd As IntPtr,
                                             <[In]> lptpm As IntPtr) As Integer
    End Function

    <DllImport("User32.dll", EntryPoint:="AppendMenuW")>
    <PreserveSig> Private Shared Function AppendMenuW(<[In]> hMenu As IntPtr,
                                                      <[In]> uFlags As MF_FLAGS,
                                                      <[In]> uIDNewItem As Integer,
                                                      <[In]> lpNewItem As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("User32.dll", EntryPoint:="InsertMenuItemW")>
    <PreserveSig> Private Shared Function InsertMenuItemW(<[In]> hMenu As IntPtr,
                                                          <[In]> item As Integer,
                                                          <[In], MarshalAs(UnmanagedType.Bool)> fByPosition As Boolean,
                                                          <[In]> ByRef lpmi As MENUITEMINFOW) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("User32.dll", EntryPoint:="GetMenuStringW")>
    <PreserveSig> Private Shared Function GetMenuStringW(<[In]> hMenu As IntPtr,
                                                         <[In]> uIDItem As Integer,
                                                         <Out, MarshalAs(UnmanagedType.LPWStr)> lpString As StringBuilder,
                                                         <[In]> cchMax As Integer,
                                                         <[In]> flags As Integer) As Integer
    End Function

    <DllImport("User32.dll", EntryPoint:="DestroyMenu")>
    Private Shared Function DestroyMenu(<[In]> hMenu As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
    End Function

    <DllImport("User32.dll", EntryPoint:="GetKeyState")>
    Private Shared Function GetKeyState(<[In]> nVirtKey As Integer) As Short
    End Function

    <DllImport("User32.dll", EntryPoint:="SetWindowLongW")>
    Private Shared Function SetWindowLong(<[In]> hWnd As IntPtr,
                                          <[In]> nIndex As Integer,
                                          <[In]> dwNewLong As SubClassDelegate) As SubClassDelegate
    End Function

    <DllImport("User32.dll", EntryPoint:="CallWindowProcW")>
    Private Shared Function CallWindowProc(<[In]> lpPrevWndFunc As SubClassDelegate,
                                           <[In]> hwnd As IntPtr,
                                           <[In]> msg As Integer,
                                           <[In]> wParam As Integer,
                                           <[In]> lParam As IntPtr) As Integer
    End Function
#End Region

#Region "Enum"
    Private Enum MF_FLAGS As Integer
        MF_STRING = &H0
        MF_ENABLED = &H0
        MF_UNCHECKED = &H0
        MF_GRAYED = &H1
        MF_DISABLED = &H2
        MF_BITMAP = &H4
        MF_CHECKED = &H8
        MF_POPUP = &H10
        MF_MENUBARBREAK = &H20
        MF_MENUBREAK = &H40
        MF_OWNERDRAW = &H100
        MF_SEPARATOR = &H800
    End Enum

    Private Enum MIIM_FLAGS As Integer
        MIIM_STATE = &H1
        MIIM_ID = &H2
        MIIM_SUBMENU = &H4
        MIIM_CHECKMARKS = &H8
        MIIM_TYPE = &H10
        MIIM_DATA = &H20
        MIIM_STRING = &H40
        MIIM_BITMAP = &H80
        MIIM_FTYPE = &H100
    End Enum

    Private Enum MFT_FLAGS As Integer
        MFT_STRING = &H0
        MFT_BITMAP = &H4
        MFT_MENUBARBREAK = &H20
        MFT_MENUBREAK = &H40
        MFT_OWNERDRAW = &H100
        MFT_RADIOCHECK = &H200
        MFT_SEPARATOR = &H800
        MFT_RIGHTORDER = &H2000
        MFT_RIGHTJUSTIFY = &H4000
    End Enum

    Private Enum MFS_FLAGS As Integer
        MFS_ENABLED = &H0
        MFS_UNHILITE = &H0
        MFS_UNCHECKED = &H0
        MFS_GRAYED = &H3
        MFS_DISABLED = &H3
        MFS_CHECKED = &H8
        MFS_HILITE = &H80
        MFS_DEFAULT = &H1000
    End Enum

    Private Enum CMF_FLAGS As Integer
        CMF_NORMAL = &H0
        CMF_DEFAULTONLY = &H1
        CMF_VERBSONLY = &H2
        CMF_EXPLORE = &H4
        CMF_NOVERBS = &H8
        CMF_CANRENAME = &H10
        CMF_NODEFAULT = &H20
        CMF_INCLUDESTATIC = &H40
        CMF_ITEMMENU = &H80
        CMF_EXTENDEDVERBS = &H100
        CMF_DISABLEDVERBS = &H200
        CMF_ASYNCVERBSTATE = &H400
        CMF_OPTIMIZEFORINVOKE = &H800
        CMF_SYNCCASCADEMENU = &H1000
        CMF_DONOTPICKDEFAULT = &H2000
    End Enum

    Private Enum CMIC_MASK_FLAGS As Integer
        CMIC_MASK_ICON = &H10
        CMIC_MASK_HOTKEY = &H20
        CMIC_MASK_NOASYNC = &H100
        CMIC_MASK_FLAG_NO_UI = &H400
        CMIC_MASK_UNICODE = &H4000
        CMIC_MASK_NO_CONSOLE = &H8000
        CMIC_MASK_ASYNCOK = &H100000
        CMIC_MASK_SHIFT_DOWN = &H10000000
        CMIC_MASK_CONTROL_DOWN = &H40000000
        CMIC_MASK_FLAG_LOG_USAGE = &H4000000
        CMIC_MASK_NOZONECHECKS = &H800000
        CMIC_MASK_PTINVOKE = &H20000000
    End Enum

    Private Enum GCS_FLAGS As Integer
        GCS_VERBA = &H0
        GCS_HELPTEXTA = &H1
        GCS_VALIDATEA = &H2
        GCS_VERBW = &H4
        GCS_HELPTEXTW = &H5
        GCS_VALIDATEW = &H6
    End Enum

#End Region

#Region "Const"
    Private Const S_OK As Integer = 0
    Private Const MAX_PATH As Integer = 260
    Private Const VK_SHIFT As Integer = &H10
    Private Const VK_CONTROL As Integer = &H11
    Private Const WM_MENUSELECT As Integer = &H11F
    Private Const CMD_PRE_CONTEXT_MENU_FIRST As Integer = &H1
    Private Const CMD_PRE_CONTEXT_MENU_LAST As Integer = &H63
    Private Const CMD_CONTEXT_MENU_FIRST As Integer = &H64
    Private Const CMD_CONTEXT_MENU_LAST As Integer = &H7FFF
    Private Const CMD_POST_CONTEXT_MENU_FIRST As Integer = &H8000
    Private Const CMD_POST_CONTEXT_MENU_LAST As Integer = &HFFFE
    Private Const GWL_WNDPROC As Integer = -4
    Private Const TPM_RETURNCMD As Integer = &H100
    Private Const SW_SHOWDEFAULT As Integer = &HA
    Private Const BHID_SFUIObject As String = "3981e225-f559-11d3-8e3a-00c04f6837d5"
    Private Const IID_IShellFolder As String = "000214e6-0000-0000-c000-000000000046"
    Private Const IID_IContextMenu As String = "000214e4-0000-0000-c000-000000000046"
    Private Const IID_IContextMenu2 As String = "000214f4-0000-0000-c000-000000000046"
    Private Const IID_IContextMenu3 As String = "bcfce0a0-ec17-11d0-8d10-00a0c90f2719"
    Private Const IID_IShellItemArray As String = "b63ea76d-1f85-456f-a19c-48159efa858b"
#End Region

#Region "Events"
    Public Event MenuCreated(MenuHandle As IntPtr)
    Public Event MenuDestroyed()
    Public Event MenuItem(MenuItemID As Integer, MenuItemString As String, MenuHandle As IntPtr)
    Public Event MenuItemReturn(MenuItemID As Integer, CommandString As String, HelpTextString As String)
#End Region

#Region "Variable"
    Private m_OldWinProc As SubClassDelegate
    Private m_IContextMenu2 As Object = Nothing
    Private m_IContextMenu3 As Object = Nothing
#End Region

#Region "Structure"
    <StructLayout(LayoutKind.Sequential)>
    Private Structure CMINVOKECOMMANDINFOEX
        Dim cbSize As Integer
        Dim fMask As CMIC_MASK_FLAGS
        Dim hwnd As IntPtr
        Dim lpVerb As IntPtr
        Dim lpParameters As IntPtr
        Dim lpDirectory As IntPtr
        Dim nShow As Integer
        Dim dwHotKey As Integer
        Dim hIcon As IntPtr
        Dim lpTitle As IntPtr
        Dim lpVerbW As IntPtr
        Dim lpParametersW As IntPtr
        Dim lpDirectoryW As IntPtr
        Dim lpTitleW As IntPtr
        Dim ptInvokeX As Integer
        Dim ptInvokeY As Integer
        Sub New(StructSize As Integer)
            cbSize = StructSize
        End Sub
    End Structure

    <StructLayout(LayoutKind.Sequential)>
    Private Structure MENUITEMINFOW
        Dim cbSize As Integer
        Dim fMask As MIIM_FLAGS
        Dim fType As MFT_FLAGS
        Dim fState As MFS_FLAGS
        Dim wID As Integer
        Dim hSubMenu As IntPtr
        Dim hbmpChecked As IntPtr
        Dim hbmpUnchecked As IntPtr
        Dim dwItemData As Integer
        Dim dwTypeData As IntPtr
        Dim cch As Integer
        Dim hbmpItem As IntPtr
        Sub New(StructSize As Integer)
            cbSize = StructSize
        End Sub
    End Structure
#End Region

#Region "Subclass Owner"
    Private ReadOnly WinProc As New SubClassDelegate(AddressOf WinProcFunc)

    Private Delegate Function SubClassDelegate(<[In]> hwnd As IntPtr,
                                               <[In]> msg As Integer,
                                               <[In]> wParam As Integer,
                                               <[In]> lParam As IntPtr) As Integer

    Private Function WinProcFunc(hWnd As IntPtr, uMsg As Integer, wParam As Integer, lParam As IntPtr) As Integer

        ' ****************************** kleine Spielerei ******************************
        'Optional: Eigenschaften des MenüItems ermitteln
        If uMsg = WM_MENUSELECT AndAlso lParam <> IntPtr.Zero Then

            Dim sbMenuStr As New StringBuilder(256)
            GetMenuStringW(lParam, LoWord(wParam), sbMenuStr, sbMenuStr.Capacity, 400)
            RaiseEvent MenuItem(LoWord(wParam), sbMenuStr.ToString, lParam)

        End If
        ' ******************************************************************************

        If m_IContextMenu3 IsNot Nothing Then

            Dim intRet As Integer

            If CType(m_IContextMenu3, IContextMenu3).HandleMenuMsg2(uMsg, wParam, lParam, intRet) = S_OK Then
                Return intRet
            End If

        ElseIf m_IContextMenu2 IsNot Nothing Then

            If CType(m_IContextMenu2, IContextMenu2).HandleMenuMsg(uMsg, wParam, lParam) = S_OK Then
                Return 0
            End If

        End If

        Return CallWindowProc(m_OldWinProc, hWnd, uMsg, wParam, lParam)

    End Function
#End Region

#Region "Public Functions"
    Public Function ShowContextMenu(FolderFileList As List(Of String),
                                    Owner As IWin32Window,
                                    Optional xMenuPos As Integer = 0,
                                    Optional YMenuPos As Integer = 0,
                                    Optional BgMenu As Boolean = False) As Boolean

        ' Funktion verlassen wenn Bedingungen erfüllt sind
        If Owner Is Nothing Then Return False
        If FolderFileList Is Nothing Then Return False

        Dim eCMFF As CMF_FLAGS
        Dim bolRet As Boolean = False
        Dim oIContextMenu As Object = Nothing

        ' ein IContextMenu erstellen
        If CreateShellContextMenu(Owner.Handle, FolderFileList, BgMenu, oIContextMenu, eCMFF) Then

            If oIContextMenu IsNot Nothing Then

                Dim pIContextMenu2 As IntPtr = IntPtr.Zero

                ' ein IContextMenu2 erstellen
                If Marshal.QueryInterface(Marshal.GetIUnknownForObject(oIContextMenu), New Guid(IID_IContextMenu2), pIContextMenu2) = S_OK Then

                    If pIContextMenu2 <> IntPtr.Zero Then

                        Dim pIContextMenu3 As IntPtr = IntPtr.Zero

                        ' ein IContextMenu3 erstellen
                        If Marshal.QueryInterface(pIContextMenu2, New Guid(IID_IContextMenu3), pIContextMenu3) = S_OK Then

                            If pIContextMenu3 <> IntPtr.Zero Then

                                m_IContextMenu3 = Marshal.GetObjectForIUnknown(pIContextMenu3)
                                Marshal.Release(pIContextMenu3)

                            End If

                        End If

                        m_IContextMenu2 = Marshal.GetObjectForIUnknown(pIContextMenu2)
                        Marshal.Release(pIContextMenu2)

                    End If

                End If

                ' Subclassing des Owner beginnen
                If m_OldWinProc Is Nothing Then
                    m_OldWinProc = SetWindowLong(Owner.Handle, GWL_WNDPROC, WinProc)
                End If

                ' ein Popupmenü erstellen
                Dim hMenu As IntPtr = CreatePopupMenu()

                If hMenu <> IntPtr.Zero Then

                    RaiseEvent MenuCreated(hMenu)

                    Dim eCMICMF As CMIC_MASK_FLAGS = CMIC_MASK_FLAGS.CMIC_MASK_ASYNCOK

                    ' Shift gedrückt?
                    If CBool(GetKeyState(VK_SHIFT) And &H8000) Then

                        eCMFF = eCMFF Or CMF_FLAGS.CMF_EXTENDEDVERBS
                        eCMICMF = eCMICMF Or CMIC_MASK_FLAGS.CMIC_MASK_SHIFT_DOWN

                    End If

                    ' Ctrl/Strg gedrückt?
                    If CBool(GetKeyState(VK_CONTROL) And &H8000) Then

                        eCMFF = eCMFF Or CMF_FLAGS.CMF_EXTENDEDVERBS
                        eCMICMF = eCMICMF Or CMIC_MASK_FLAGS.CMIC_MASK_CONTROL_DOWN

                    End If

                    ' Kontextmenü laden
                    If CType(oIContextMenu, IContextMenu).QueryContextMenu(hMenu, 0, CMD_CONTEXT_MENU_FIRST, CMD_CONTEXT_MENU_LAST, eCMFF) > 0 Then

                        ' ****************************** kleine Spielerei ******************************
                        ' Optional: Eigene MenuItems/SubMenüs usw hinzufügen
                        AddPreMenuItems(hMenu)
                        AddPostMenuItems(hMenu)
                        ' ******************************************************************************

                        ' Kontextmenü anzeigen
                        Dim retCmd As Integer = TrackPopupMenuEx(hMenu, TPM_RETURNCMD, xMenuPos, YMenuPos, Owner.Handle, IntPtr.Zero)

                        If retCmd > CMD_CONTEXT_MENU_FIRST AndAlso retCmd <= CMD_CONTEXT_MENU_LAST Then

                            ' Ab hier könnte man auf das MenuItem reagieren das durch die Shell erzeugt wurde.
                            ' InvokeCommand dann halt nicht ausführen und dafür was eigenes implementieren.

                            Dim sbCommandString As New StringBuilder(MAX_PATH)
                            CType(oIContextMenu, IContextMenu).GetCommandString(New IntPtr(retCmd - CMD_CONTEXT_MENU_FIRST),
                                                                                GCS_FLAGS.GCS_VERBA, IntPtr.Zero,
                                                                                sbCommandString, sbCommandString.Capacity)

                            Dim sbHelpTextString As New StringBuilder(MAX_PATH)
                            CType(oIContextMenu, IContextMenu).GetCommandString(New IntPtr(retCmd - CMD_CONTEXT_MENU_FIRST),
                                                                            GCS_FLAGS.GCS_HELPTEXTA, IntPtr.Zero,
                                                                            sbHelpTextString, sbHelpTextString.Capacity)

                            RaiseEvent MenuItemReturn(retCmd, sbCommandString.ToString, sbHelpTextString.ToString)

                            Dim tCMICI As New CMINVOKECOMMANDINFOEX(Marshal.SizeOf(Of CMINVOKECOMMANDINFOEX))

                            ' Befehl, der aus dem Kontextmenü ausgeführt werden soll
                            With tCMICI
                                .lpVerb = New IntPtr(retCmd - CMD_CONTEXT_MENU_FIRST)
                                .hwnd = Owner.Handle
                                .fMask = eCMICMF
                                .nShow = SW_SHOWDEFAULT
                            End With

                            ' Befehl ausführen
                            If CType(oIContextMenu, IContextMenu).InvokeCommand(tCMICI) = S_OK Then bolRet = True

                        Else

                            If retCmd = 0 Then
                                ' Menü hat sich geschlossen ohne das ein MenuItem ausgewählt wurde
                                RaiseEvent MenuItemReturn(retCmd, String.Empty, String.Empty)
                            Else
                                'hier können selbst hinzugefügte MenüItems ausgewertet werden
                                RaiseEvent MenuItemReturn(retCmd, String.Empty, String.Empty)
                            End If

                        End If

                    End If

                    ' Popupmenü löschen
                    If DestroyMenu(hMenu) Then RaiseEvent MenuDestroyed()

                End If

                ' Owner Subclasing beenden
                If m_OldWinProc IsNot Nothing Then
                    If SetWindowLong(Owner.Handle, GWL_WNDPROC, m_OldWinProc) IsNot Nothing Then m_OldWinProc = Nothing
                End If

                ' IContextMenu3 löschen
                If m_IContextMenu3 IsNot Nothing Then Marshal.ReleaseComObject(m_IContextMenu3)

                ' IContextMenu2 löschen
                If m_IContextMenu2 IsNot Nothing Then Marshal.ReleaseComObject(m_IContextMenu2)

                ' IContextMenu löschen
                Marshal.FinalReleaseComObject(oIContextMenu)

            End If

        End If

        Return bolRet

    End Function
#End Region

#Region "Private Functions"

    Private Function CreateShellContextMenu(OwnerHwnd As IntPtr,
                                            FolderFileList As List(Of String),
                                            BgMenu As Boolean,
                                            ByRef oIContextMenu As Object,
                                            ByRef eCMFF As CMF_FLAGS) As Boolean
        Dim bolRet As Boolean = False

        If FolderFileList.Count > 0 Then

            If BgMenu Then

                Dim pPIDL As IntPtr = IntPtr.Zero

                ' nur vom ersten Eintrag in der Liste eine PIDL ermitteln
                If SHParseDisplayName(FolderFileList(0), IntPtr.Zero, pPIDL,
                                      IntPtr.Zero, IntPtr.Zero) = S_OK Then

                    If pPIDL <> IntPtr.Zero Then

                        Dim pPIDLChild As IntPtr = IntPtr.Zero
                        Dim oISHellFolder As Object = Nothing

                        ' von der PIDL ein IShellFolder und eine weitere PIDL erstellen
                        If SHBindToParent(pPIDL, New Guid(IID_IShellFolder),
                                          oISHellFolder, pPIDLChild) = S_OK Then

                            If pPIDLChild <> IntPtr.Zero Then

                                If oISHellFolder IsNot Nothing Then

                                    Dim oISHellFolder_2 As Object = Nothing

                                    ' vom IShellFolder und der PIDL einen weiteres IShellFolder erstellen
                                    If CType(oISHellFolder, IShellFolder).BindToObject(pPIDLChild, IntPtr.Zero,
                                                                                       New Guid(IID_IShellFolder),
                                                                                       oISHellFolder_2) = S_OK Then

                                        If oISHellFolder_2 IsNot Nothing Then

                                            ' vom IShellFolder ein IContextMenu erstellen
                                            If CType(oISHellFolder_2, IShellFolder).CreateViewObject(OwnerHwnd,
                                                                                                     New Guid(IID_IContextMenu),
                                                                                                     oIContextMenu) = S_OK Then

                                                eCMFF = CMF_FLAGS.CMF_NORMAL

                                                bolRet = True

                                            End If

                                            Marshal.ReleaseComObject(oISHellFolder_2)

                                        End If

                                    End If

                                    Marshal.ReleaseComObject(oISHellFolder)

                                End If

                            End If

                        End If

                        Marshal.FreeCoTaskMem(pPIDL)

                    End If

                End If

            Else

                Dim pPidlList As New List(Of IntPtr)

                ' alle Einträge in der Liste durchlaufen und die PIDLs in eine Liste speichern
                For Each strFolderFile As String In FolderFileList

                    Dim pPIDL As IntPtr = IntPtr.Zero

                    If SHParseDisplayName(strFolderFile, IntPtr.Zero, pPIDL,
                                          IntPtr.Zero, IntPtr.Zero) = S_OK Then

                        pPidlList.Add(pPIDL)

                    End If

                Next

                If pPidlList.Count > 0 Then

                    Dim IIShellItemArray As IShellItemArray = Nothing

                    ' vom PIDL-Array ein IShellItemArray erstellen
                    If SHCreateShellItemArrayFromIDLists(pPidlList.Count, pPidlList.ToArray,
                                                         IIShellItemArray) = S_OK Then

                        If IIShellItemArray IsNot Nothing Then

                            ' vom IShellItemArray ein IContextMenu erstellen
                            If IIShellItemArray.BindToHandler(IntPtr.Zero, New Guid(BHID_SFUIObject),
                                                              New Guid(IID_IContextMenu), oIContextMenu) = S_OK Then

                                eCMFF = CMF_FLAGS.CMF_NORMAL Or CMF_FLAGS.CMF_CANRENAME

                                bolRet = True

                            End If

                            Marshal.FinalReleaseComObject(IIShellItemArray)

                        End If

                    End If

                    For Each pPidl As IntPtr In pPidlList
                        Marshal.FreeCoTaskMem(pPidl)
                    Next

                End If

            End If

        End If

        Return bolRet

    End Function

    Private Function LoWord(number As Integer) As Integer
        Return number And &HFFFF
    End Function

    ' ****************************** kleine Spielerei ******************************
    Private Sub AddPreMenuItems(hMenu As IntPtr)
        'Optional: fügt Menüitems vor, mittendrin oder am Ende eines Menüs hinzu

        Dim tMENUITEMINFOW As New MENUITEMINFOW(Marshal.SizeOf(Of MENUITEMINFOW))

        Using MenuBmp As New Bitmap(16, 16, PixelFormat.Format32bppArgb)
            Using gra As Graphics = Graphics.FromImage(MenuBmp)
                gra.FillRectangle(Brushes.Red, 0, 0, 16, 16)
            End Using
            With tMENUITEMINFOW
                .fMask = MIIM_FLAGS.MIIM_STRING Or MIIM_FLAGS.MIIM_BITMAP Or
                         MIIM_FLAGS.MIIM_STATE Or MIIM_FLAGS.MIIM_ID Or MIIM_FLAGS.MIIM_FTYPE
                .fType = MFT_FLAGS.MFT_STRING
                .fState = MFS_FLAGS.MFS_ENABLED
                .dwTypeData = Marshal.StringToHGlobalUni("Pre MenuItem 1")
                .wID = CMD_PRE_CONTEXT_MENU_FIRST
                .hbmpItem = MenuBmp.GetHbitmap
            End With
            InsertMenuItemW(hMenu, CMD_PRE_CONTEXT_MENU_FIRST - 1, True, tMENUITEMINFOW)

            Using gra As Graphics = Graphics.FromImage(MenuBmp)
                gra.FillRectangle(Brushes.Blue, 0, 0, 16, 16)
            End Using
            With tMENUITEMINFOW
                .fMask = MIIM_FLAGS.MIIM_STRING Or MIIM_FLAGS.MIIM_BITMAP Or
                         MIIM_FLAGS.MIIM_STATE Or MIIM_FLAGS.MIIM_ID Or MIIM_FLAGS.MIIM_FTYPE
                .fType = MFT_FLAGS.MFT_STRING
                .fState = MFS_FLAGS.MFS_ENABLED
                .dwTypeData = Marshal.StringToHGlobalUni("Pre MenuItem 2")
                .wID = CMD_PRE_CONTEXT_MENU_FIRST + 1
                .hbmpItem = MenuBmp.GetHbitmap
            End With
            InsertMenuItemW(hMenu, CMD_PRE_CONTEXT_MENU_FIRST, True, tMENUITEMINFOW)
        End Using

        With tMENUITEMINFOW
            .fMask = MIIM_FLAGS.MIIM_STATE Or MIIM_FLAGS.MIIM_ID Or MIIM_FLAGS.MIIM_FTYPE
            .fType = MFT_FLAGS.MFT_SEPARATOR
            .fState = MFS_FLAGS.MFS_ENABLED
            .wID = CMD_PRE_CONTEXT_MENU_FIRST + 2
        End With
        InsertMenuItemW(hMenu, CMD_PRE_CONTEXT_MENU_FIRST + 1, True, tMENUITEMINFOW)
    End Sub

    Private Sub AddPostMenuItems(hMenu As IntPtr)
        ' Optional: fügt Menüitems nur am Ende eines Menüs hinzu
        AppendMenuW(hMenu, MF_FLAGS.MF_SEPARATOR, CMD_POST_CONTEXT_MENU_FIRST, Nothing)
        AppendMenuW(hMenu, MF_FLAGS.MF_STRING, CMD_POST_CONTEXT_MENU_FIRST + 1, Marshal.StringToHGlobalUni("Post MenuItem 1"))
        AppendMenuW(hMenu, MF_FLAGS.MF_STRING, CMD_POST_CONTEXT_MENU_FIRST + 2, Marshal.StringToHGlobalUni("Post MenuItem 2"))
    End Sub
    ' ******************************************************************************
#End Region

#Region "Interface IShellFolder"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IShellFolder)>
    Private Interface IShellFolder
        <PreserveSig> Function ParseDisplayName(<[In]> hwnd As IntPtr,
                                                <[In]> pbc As IntPtr,
                                                <[In], MarshalAs(UnmanagedType.LPWStr)> pszDisplayName As String,
                                                <[In], Out> ByRef pchEaten As Integer,
                                                <Out> ByRef ppidl As IntPtr,
                                                <[In], Out> ByRef pdwAttributes As Integer) As Integer
        <PreserveSig> Function EnumObjects() As Integer ' Dummy
        <PreserveSig> Function BindToObject(<[In]> pidl As IntPtr,
                                            <[In]> pbc As IntPtr,
                                            <[In]> ByRef riid As Guid,
                                            <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppv As Object) As Integer
        <PreserveSig> Function BindToStorage() As Integer ' Dummy
        <PreserveSig> Function CompareIDs() As Integer ' Dummy
        <PreserveSig> Function CreateViewObject(<[In]> hwnd As IntPtr,
                                                <[In]> ByRef riid As Guid,
                                                <Out, MarshalAs(UnmanagedType.Interface)> ByRef ppv As Object) As Integer
        <PreserveSig> Function GetAttributesOf() As Integer ' Dummy
        <PreserveSig> Function GetUIObjectOf() As Integer ' Dummy
        <PreserveSig> Function GetDisplayNameOf() As Integer ' Dummy
        <PreserveSig> Function SetNameOf() As Integer ' Dummy

    End Interface
#End Region

#Region "Interface IContextMenu"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IContextMenu)>
    Private Interface IContextMenu
        <PreserveSig> Function QueryContextMenu(<[In]> hMenu As IntPtr,
                                                <[In]> indexMenu As Integer,
                                                <[In]> idCmdFirst As Integer,
                                                <[In]> idCmdLast As Integer,
                                                <[In]> uFlags As CMF_FLAGS) As Integer
        <PreserveSig> Function InvokeCommand(<[In]> ByRef pici As CMINVOKECOMMANDINFOEX) As Integer
        <PreserveSig> Function GetCommandString(<[In]> idCmd As IntPtr,
                                                <[In]> uType As GCS_FLAGS,
                                                <[In]> pReserved As IntPtr,
                                                <[In], Out, MarshalAs(UnmanagedType.LPStr)> pszName As StringBuilder,
                                                <[In]> cchMax As Integer) As Integer

    End Interface
#End Region

#Region "Interface IContextMenu2"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IContextMenu2)>
    Private Interface IContextMenu2
        Inherits IContextMenu
        <PreserveSig> Shadows Function QueryContextMenu(<[In]> hMenu As IntPtr,
                                                        <[In]> indexMenu As Integer,
                                                        <[In]> idCmdFirst As Integer,
                                                        <[In]> idCmdLast As Integer,
                                                        <[In]> uFlags As CMF_FLAGS) As Integer
        <PreserveSig> Shadows Function InvokeCommand(<[In]> ByRef pici As CMINVOKECOMMANDINFOEX) As Integer
        <PreserveSig> Shadows Function GetCommandString(<[In]> idCmd As IntPtr,
                                                        <[In]> uType As GCS_FLAGS,
                                                        <[In]> pReserved As IntPtr,
                                                        <[In], Out, MarshalAs(UnmanagedType.LPStr)> pszName As StringBuilder,
                                                        <[In]> cchMax As Integer) As Integer
        <PreserveSig> Function HandleMenuMsg(<[In]> uMsg As Integer,
                                             <[In]> wParam As Integer,
                                             <[In]> lParam As IntPtr) As Integer
    End Interface
#End Region

#Region "Interface IContextMenu3"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IContextMenu3)>
    Private Interface IContextMenu3
        Inherits IContextMenu2
        <PreserveSig> Shadows Function QueryContextMenu(<[In]> hMenu As IntPtr,
                                                        <[In]> indexMenu As Integer,
                                                        <[In]> idCmdFirst As Integer,
                                                        <[In]> idCmdLast As Integer,
                                                        <[In]> uFlags As CMF_FLAGS) As Integer
        <PreserveSig> Shadows Function InvokeCommand(<[In]> ByRef pici As CMINVOKECOMMANDINFOEX) As Integer
        <PreserveSig> Shadows Function GetCommandString(<[In]> idCmd As IntPtr,
                                                        <[In]> uType As GCS_FLAGS,
                                                        <[In]> pReserved As IntPtr,
                                                        <[In], Out, MarshalAs(UnmanagedType.LPStr)> pszName As StringBuilder,
                                                        <[In]> cchMax As Integer) As Integer
        <PreserveSig> Shadows Function HandleMenuMsg(<[In]> uMsg As Integer,
                                                     <[In]> wParam As Integer,
                                                     <[In]> lParam As IntPtr) As Integer
        <PreserveSig> Function HandleMenuMsg2(<[In]> uMsg As Integer,
                                              <[In]> wParam As Integer,
                                              <[In]> lParam As IntPtr,
                                              <Out> ByRef plResult As Integer) As Integer
    End Interface
#End Region

#Region "Interface IShellItemArray"
    <ComImport>
    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown)>
    <Guid(IID_IShellItemArray)>
    Private Interface IShellItemArray
        <PreserveSig> Function BindToHandler(<[In]> pbc As IntPtr,
                                             <[In]> ByRef bhid As Guid,
                                             <[In]> ByRef riid As Guid,
                                             <Out, MarshalAs(UnmanagedType.IUnknown)> ByRef ppvOut As Object) As Integer
        'GetPropertyStore
        'GetPropertyDescriptionList
        'GetAttributes
        'GetCount
        'GetItemAt
        'EnumItems
    End Interface
#End Region

End Class
